(**************************************************************************)
(*                                                                        *)
(*  This file was originally part of the Objective Caml Library           *)
(*                                                                        *)
(*                           Objective Caml                               *)
(*                                                                        *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt            *)
(*                                                                        *)
(*  Copyright 1996 Institut National de Recherche en Informatique et      *)
(*  en Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the GNU Library General Public License, with       *)
(*  the special exception on linking described in file                    *)
(*  licenses/CAML_LICENSE                                                 *)
(*                                                                        *)
(*  File modified by CEA (Commissariat  l'nergie Atomique)              *)
(**************************************************************************)

(* $Id: myDynlink.mli,v 1.2 2008/08/28 09:22:49 uid528 Exp $ *)

module type OldDynlink = sig

  (** {6 Dynamic loading of compiled files} *)

val loadfile : string -> unit
(** In bytecode: load the given bytecode object file ([.cmo] file) or
    bytecode library file ([.cma] file), and link it with the running 
    program. In native code: load the given OCaml plugin file (usually
    [.cmxs]), and link it with the running 
    program.
    All toplevel expressions in the loaded compilation units
    are evaluated. No facilities are provided to
    access value names defined by the unit. Therefore, the unit
    must register itself its entry points with the main program,
    e.g. by modifying tables of functions. *)

(** {6 Access control} *)

val allow_unsafe_modules : bool -> unit
(** Govern whether unsafe object files are allowed to be
    dynamically linked. A compilation unit is ``unsafe'' if it contains
    declarations of external functions, which can break type safety.
    By default, dynamic linking of unsafe object files is
    not allowed. In native code, this function does nothing; object files
    with external functions are always allowed to be dynamically linked. *)

(** {6 Deprecated, initialization} *)

val init : unit -> unit
(** @deprecated Initialize the [Dynlink] library. This function is called
    automatically when needed. *)

type linking_error = Dynlink.linking_error =
    Undefined_global of string
  | Unavailable_primitive of string
  | Uninitialized_global of string


(**/**)

(** {6 Internal functions} *)

val digest_interface : string -> string list -> Digest.t

end

include OldDynlink

exception Unsupported_Feature of string

(** Dynamic loading of object files. *)

val is_native: bool
(** [true] if the program is native,
    [false] if the program is bytecode. *)

(** {6 Error reporting} *)

type error =
    Not_a_bytecode_file of string
  | Inconsistent_import of string
  | Unavailable_unit of string
  | Unsafe_file
  | Linking_error of string * linking_error
  | Corrupted_interface of string
  | File_not_found of string
  | Cannot_open_dll of string
  | Inconsistent_implementation of string

exception Error of error
(** Errors in dynamic linking are reported by raising the [Error]
    exception with a description of the error. *)

val error_message : error -> string
(** Convert an error description to a printable message. *)

(*
Local Variables:
compile-command: "LC_ALL=C make -C ../.. -j"
End:
*)
